freeslot(
"MT_EGGS_HOMINGBOMB",
"MT_EGGS_HOMINGBOMB_GLOW",
"S_EGGS_HOMINGBOMB_WANDER",
"S_EGGS_HOMINGBOMB_CHASE",
"S_EGGS_HOMINGBOMB_HUG",
"S_EGGS_HOMINGBOMB_SPAWN",
"S_EGGS_HOMINGBOMB_EXPLODE",
"SPR_AMBM",
"sfx_bmltch",
"MT_EGGS_TAILSDOLL"
)

mobjinfo[MT_EGGS_HOMINGBOMB] = {
	spawnstate = S_EGGS_HOMINGBOMB_SPAWN,
	speed = 32*FRACUNIT,
	radius = 16*FRACUNIT,
	height = 32*FRACUNIT,
	seesound = sfx_ghosty,
	deathsound = sfx_brakrx,
	deathstate = S_EGGS_HOMINGBOMB_EXPLODE,
	xdeathstate = S_EGGS_HOMINGBOMB_EXPLODE,
	seestate = S_EGGS_HOMINGBOMB_WANDER,
	dispoffset = 2
}

mobjinfo[MT_EGGS_HOMINGBOMB_GLOW] = {
	spawnstate = S_INVISIBLE,
	flags = MF_NOBLOCKMAP|MF_NOCLIPTHING|MF_NOGRAVITY|MF_NOCLIP|MF_NOCLIPHEIGHT,
	dispoffset = 6
}

states[S_EGGS_HOMINGBOMB_SPAWN] = {SPR_AMBM, E, 1, A_MinusCheck, 0, 0, S_EGGS_HOMINGBOMB_SPAWN}
states[S_EGGS_HOMINGBOMB_WANDER] = {SPR_AMBM, F|FF_ANIMATE, -1, A_AmyBombWander, 3, 3, S_EGGS_HOMINGBOMB_WANDER}
states[S_EGGS_HOMINGBOMB_CHASE] = {SPR_AMBM, F|FF_ANIMATE, -1, A_AmyBombChase, 3, 3, S_EGGS_HOMINGBOMB_CHASE}
states[S_EGGS_HOMINGBOMB_HUG] = {SPR_AMBM, B, -1, nil, 0, 0, S_EGGS_HOMINGBOMB_HUG}

states[S_EGGS_HOMINGBOMB_EXPLODE] = {SPR_BARX, A|FF_ANIMATE, 12, nil, 3, 3, S_NULL}

local function TeamColorName(str, player)
	
	if (G_GametypeHasTeams()) then
		if (player.ctfteam == 1) then
			str = "\x85"..str.."\x80"
		else
			str = "\x84"..str.."\x80"
		end
	end
	
	return str
end

//source of the radius attack is attributed to the amebomb instead of the player
//this way it'll still damage players
//
//note that this means it wont award the player points for defeating enemies
//with it however
local function RealExplosion(mo)
	//reuse fadein object as the source for the explosion sound to prevent cutoff
	mo.fadein.state = S_INVISIBLE
	S_StartSound(mo.fadein, sfx_brakrx)
	mo.fadein.fuse = TICRATE
		
	mo.momx,mo.momy,mo.momz = 0,0,0
	mo.flags = MF_NOCLIP|MF_NOGRAVITY

	P_RadiusAttack(mo, mo, 128*FRACUNIT, 5, true)
end

addHook("MobjFuse", RealExplosion, MT_EGGS_HOMINGBOMB)

//note: this hook runs twice on damaging hits
//ends at the first if though, thanks to the player now being the source
local function BombFriendlyFire(pmo, bomb, src)


	if not (pmo.player) then return end

	//not a bomb
	if not (bomb)
	or not (bomb.valid)
	or not (bomb.type == MT_EGGS_HOMINGBOMB) then
		return
	end

	//if (bomb.team == pmo.player.ctfteam)
	if (G_GametypeHasTeams())
	and (CBW_Battle.MyTeam(bomb, pmo.player))
		return false
	end
	

	if (bomb.dodamage)
		bomb.dodamage = false
		return
	end
	
	if (bomb.setter == pmo) then
		return false
	end

	bomb.dodamage = true
	pmo.amybombgrabbed = false
	P_DamageMobj(pmo, bomb, bomb.target)
	return false
end

addHook("ShouldDamage", BombFriendlyFire, MT_PLAYER)

//rewrite of P_LookForPlayers to ignore specific players
local function BombSearch(actor)

	local dist = 512*FRACUNIT
	
	for player in players.iterate do
	
		if (player.spectator)
			continue
		end
		
		if (G_GametypeHasTeams())
			if (actor.target)
			and (CBW_Battle.MyTeam(actor.target, player))
			or (actor.team == player.ctfteam)
				continue
			end
		end

		if (player.pflags & PF_INVIS)
			continue // ignore notarget
		end

		if not (player.mo)
			continue
		end
		
		if not (player.mo.flags & MF_SHOOTABLE) then
			continue
		end
		
		if (player.mo.amybombgrabbed) then
			continue //already has bomb attached
		end

		if (player.mo.health <= 0)
			continue // dead
		end
		
		if (actor.target == player.mo) then
			continue
		end

		if (dist > 0)
		
		and (P_AproxDistance(P_AproxDistance(player.mo.x - actor.x, player.mo.y - actor.y), player.mo.z - actor.z) > dist)
			continue // Too far away
		end

		if not (P_CheckSight(actor, player.mo))
			continue // out of sight
		end
		
		actor.tracer = player.mo

		S_StartSound(actor, mobjinfo[actor.type].seesound)
		actor.state = S_EGGS_HOMINGBOMB_CHASE
		
		return true
	end

	actor.tracer = nil
	//return false;
end

function A_AmyBombWander(mo)
	
	if (BombSearch(mo)) then
		local broked = P_SpawnMobj(mo.x, mo.y, mo.z+mo.height, MT_GHOST)
		broked.fuse = 15
		broked.state = S_ALART1
	else
		P_InstaThrust(mo, mo.angle, 8*FRACUNIT)
		mo.angle = $1 + ANG2*2
	end
end

function A_AmyBombChase(mo)
	if not (mo.tracer)
	or (mo.tracer.amybombgrabbed) then
		mo.state = S_EGGS_HOMINGBOMB_WANDER
		return
	end

	local angletoplayer = R_PointToAngle2(mo.x, mo.y, mo.tracer.x, mo.tracer.y)

	local mom,maxdist
	P_Thrust(mo, angletoplayer, (2*FRACUNIT))// >> 1)
	mom = FixedHypot(mo.momx, mo.momy);
	maxdist = 1700*mo.scale
	maxdist = P_AproxDistance(mo.x - mo.tracer.x, mo.z - mo.tracer.z, mo.y - mo.tracer.y);

	if (maxdist < (mo.radius + mo.target.radius))
	
		mo.momx = mo.tracer.player.cmomx
		mo.momy = mo.tracer.player.cmomy
		if (mo.cvmem > TICRATE)
		and not (player.exiting)
			mo.state = S_EGGS_HOMINGBOMB_WANDER
		else
			mo.tracer.amybombgrabbed = true
			mo.state = S_EGGS_HOMINGBOMB_HUG
			S_StartSoundAtVolume(mo, sfx_bmltch, 255)
			mo.flags = $1|MF_NOGRAVITY
			mo.angle = angletoplayer
			return
		end
	
	else
		maxdist = $1/3;
		if (maxdist > 30*mo.scale)
			maxdist = 30*mo.scale
		end
		if (mom > maxdist)
		and (maxdist > mo.scale)
		
			maxdist = FixedDiv(maxdist, mom);
			mo.momx = FixedMul(mo.momx, maxdist);
			mo.momy = FixedMul(mo.momy, maxdist);
		end
		
		if (abs(mo.momx) > mo.scale)
		or (abs(mo.momy) > mo.scale)
			mo.angle = R_PointToAngle2(0, 0, mo.momx, mo.momy)
		end
	end
end


///actions need to be continuously called to work
//so let's just call it in mobjthinker so FF_ANIMATE works
local function AmyBombThinker(mo)

	if (mo.state == S_EGGS_HOMINGBOMB_EXPLODE) then return end


	if (mo.state == S_EGGS_HOMINGBOMB_HUG)
		if not (mo.tracer.valid) then
			P_RemoveMobj(mo)
		end
		//spawning the bomb on someone causes it to not gain nogravity so let's force it here
		mo.flags = $1|MF_NOGRAVITY
		P_TeleportMove(mo, mo.tracer.x, mo.tracer.y, mo.tracer.z+((mo.tracer.height/3)*P_MobjFlip(mo.target)))
	end

	if (mo.fuse > 3*TICRATE) then
		if (mo.state == S_EGGS_HOMINGBOMB_WANDER)
			A_AmyBombWander(mo)
		elseif (mo.state == S_EGGS_HOMINGBOMB_CHASE)
			A_AmyBombChase(mo)
		end
	
	
	//IT'S ABOUT TO BLOW, GET OUTTA THERE
	else
	
		if (mo.fuse == 3*TICRATE) then
			S_StartSound(mo, sfx_s3ka4)
		end
		
		if not (mo.fadein) then
			mo.fadein = P_SpawnMobj(mo.x, mo.y, mo.z, MT_EGGS_HOMINGBOMB_GLOW)
			mo.fadein.target = mo
			mo.fadein.sprite = mo.sprite
			mo.fadein.frame = D|FF_FULLBRIGHT
		end

	end
end

addHook("MobjThinker", AmyBombThinker, MT_EGGS_HOMINGBOMB)

local function BombGlowThinker(mo)
	if not (mo.target)
	or (mo.fuse) then return end
	
	local fadeFrame = D
	if not (mo.target.state == S_EGGS_HOMINGBOMB_HUG) then
		mo.target.frame = A
		fadeFrame = J
	else
		mo.target.frame = C
	end
	
	local tLevel = {
	FF_TRANS10,
	FF_TRANS20,
	FF_TRANS30,
	FF_TRANS40,
	FF_TRANS50,
	FF_TRANS60,
	FF_TRANS70,
	FF_TRANS80,
	FF_TRANS90
	}

	local tPos = -1+mo.target.fuse/(3*TICRATE/10)
	
	if (tPos <= 0) then
		mo.frame = fadeFrame|FF_FULLBRIGHT
	else
		mo.frame = fadeFrame|FF_FULLBRIGHT|tLevel[tPos]
	end
	
	mo.momx = mo.target.momx
	mo.momy = mo.momy
	mo.momz = mo.momz
	mo.angle = mo.target.angle
	
	mo.eflags = mo.target.eflags
	P_TeleportMove(mo, mo.target.x, mo.target.y, mo.target.z)
end
addHook("MobjThinker", BombGlowThinker, MT_EGGS_HOMINGBOMB_GLOW)

local function LayBomb(player)

	local xpld

	if (table.maxn(player.eggettebattle.amy) == 2)
		if (player.eggettebattle.amy[1].valid) then
			local amy = player.eggettebattle.amy[1]
			
			if (amy.eflags & MFE_VERTICALFLIP) then
				xpld = P_SpawnMobj(amy.x, amy.y, amy.z+amy.height, MT_GFZFLOWER1)
				xpld.eflags = $1|MFE_VERTICALFLIP
			else
				xpld = P_SpawnMobj(amy.x, amy.y, amy.z, MT_GFZFLOWER1)
			end
			
			xpld.flags = MF_NOGRAVITY
			S_StartSound(xpld, sfx_pop)
			xpld.state = S_XPLD1
			player.eggettebattle.amy[1].type = MT_GFZFLOWER1
			
			if (player.eggettebattle.amy[1].tracer)
			and (player.eggettebattle.amy[1].tracer.valid) then
				player.eggettebattle.amy[1].tracer.amybombgrabbed = nil
			end
			
			S_StopSound(player.eggettebattle.amy[1])
			if (player.eggettebattle.amy[1].fadein)
			and (player.eggettebattle.amy[1].fadein.valid) then
				P_RemoveMobj(player.eggettebattle.amy[1].fadein)
			end
			P_RemoveMobj(player.eggettebattle.amy[1])
		end
		
		table.remove(player.eggettebattle.amy, 1)
	end
	
	local egg
	
	if (player.mo.eflags & MFE_VERTICALFLIP) then
		egg = P_SpawnMobj(player.mo.x, player.mo.y, player.mo.z+player.mo.height, MT_EGGS_HOMINGBOMB)
		egg.eflags = $1|MFE_VERTICALFLIP
	else
		egg = P_SpawnMobj(player.mo.x, player.mo.y, player.mo.z, MT_EGGS_HOMINGBOMB)
	end
	
	if not (egg) then return end
	
	S_StartSound(egg, sfx_peww)
	egg.angle = player.mo.angle
	egg.target = player.mo
	egg.color = player.mo.color
	egg.fuse = 10*TICRATE
	
	egg.pname = TeamColorName(player.name, player)
	
	if (player.ctfteam != 0) then
		egg.color = player.skincolor
		egg.team = player.ctfteam
	end
	
	table.insert(player.eggettebattle.amy, egg)
end


//
//
// TAILS DOLL
//
//

local function PlaceFakeDork(player)

	if (table.maxn(player.eggettebattle.tailsdoll) == 2)
		if (player.eggettebattle.tailsdoll[1].valid) then
			P_KillMobj(player.eggettebattle.tailsdoll[1])
			player.eggettebattle.tailsdoll[1].state = S_XPLD1
		end
		
		table.remove(player.eggettebattle.tailsdoll, 1)
	end
	
	local egg
	
	if (player.mo.eflags & MFE_VERTICALFLIP) then
		egg = P_SpawnMobj(player.mo.x, player.mo.y, player.mo.z+player.mo.height, MT_EGGS_TAILSDOLL)
		egg.eflags = $1|MFE_VERTICALFLIP
	else
		egg = P_SpawnMobj(player.mo.x, player.mo.y, player.mo.z, MT_EGGS_TAILSDOLL)
	end
	
	if not (egg)
	or not (egg.valid) then
		return
	end
	
	egg.angle = player.mo.angle
	egg.tracer = player.mo
	egg.target = nil
	egg.color = player.mo.color
	
	egg.pname = TeamColorName(player.name, player)
	
	if (player.ctfteam != 0) then
		egg.color = player.skincolor
		egg.team = player.ctfteam
	end
	
	table.insert(player.eggettebattle.tailsdoll, egg)
end

mobjinfo[MT_EGGS_TAILSDOLL] = { 
	spawnstate = S_INVISIBLE,
	painstate = S_INVISIBLE,
	spawnhealth = 2,
	deathstate = S_XPLD1,
	radius = 16*FRACUNIT,
	height = 40*FRACUNIT,
	flags = MF_SLIDEME|MF_SHOOTABLE|MF_ENEMY|MF_SPECIAL
}

//
//
//
//
//

//sort out the tables for egg stuff, spawn overlays for friendlies
local function EggOverlays(pmo)

	local spawnTarget = false
	
	//give our friendlies a heads up
	if (consoleplayer) //dont know if this would break on dedicated servers otherwise and don't wanna check
		if (consoleplayer == pmo.player)
			spawnTarget = true
		end
	end
	
	//remove dead objects and spawn the target
	for _, valTable in pairs(pmo.player.eggettebattle) do
		local removedCount = 0
		
		for pos, val in pairs(valTable) do
			if not (val.valid) then
				table.remove(valTable, pos-removedCount)
				removedCount = $+1
			else
				val.pname = TeamColorName(pmo.player.name, pmo.player)
				if (spawnTarget) then
					
					if (val.health < 2)
					and (val.type == MT_EGGS_TAILSDOLL) then
						P_SpawnLockOn(consoleplayer, val, S_LOCKON2)
					else
						P_SpawnLockOn(consoleplayer, val, S_LOCKON4)
					end
				end
			end
		end
		
	end
end

//master function for battlemod
local function EggAttacks(pmo, doaction)

	if not (pmo) then return end
	local player = pmo.player

	if not (player.eggettebattle) then
		player.eggettebattle = {amy = {}, tailsdoll = {}}
	end

	local actionTable = {
	{text = "Tails Doll", rings = 15, ability = PlaceFakeDork, cooldown = 3*TICRATE},
	{text = "Amy Bomb", rings = 10, ability = LayBomb, cooldown = (TICRATE/2)+(2*TICRATE)}
	}
	
	local actionVal = 2
	
	//switch ability sheet on the ground
	if (P_IsObjectOnGround(pmo)) then
		actionVal = 1
	end

	player.actiontext = actionTable[actionVal].text
	player.actionrings = actionTable[actionVal].rings
		
	if not (player.actioncooldown)
	and (doaction)
	and (CBW_Battle.CanDoAction(player))
		CBW_Battle.PayRings(player, actionTable[actionVal].rings, true)
		actionTable[actionVal].ability(player)
		CBW_Battle.ApplyCooldown(player, actionTable[actionVal].cooldown)
	end
	
	EggOverlays(pmo)
end


//get rid of your toys once you leave
local function EggetteCleanup(player)

	if not (player.eggettebattle) then return end

	for _, valTable in pairs(player.eggettebattle) do
		for pos, val in pairs(valTable) do
			if (val.valid) then
				P_RemoveMobj(val)
			end
		end
		
	end
	
	player.eggettebattle = nil
end

addHook("PlayerQuit", EggetteCleanup)

//switched teams, kill everything
local function Eggwich(player)
	EggetteCleanup(player)
end
addHook("TeamSwitch", Eggwich)


//respawned, update minions
local function EggSpawnFix(player)
	if not (player.mo)
	or not (player.mo.skin == "eggette")
	or not (player.eggettebattle) then return end
	
	for pos, val in pairs(player.eggettebattle.amy) do
		if (val.valid) then
			val.target = player.mo
		end
	end
	
	for pos, val in pairs(player.eggettebattle.tailsdoll) do
		if (val.valid) then
			val.tracer = player.mo
		end
	end
end
addHook("PlayerSpawn", EggSpawnFix)

//set up battlemod & remove buildings when not eggy
local battleMod = false
local function BattleModSetup()
	if not (battleMod)
		if (CBW_Battle) then
			battleMod = true
			
			CBW_Battle.SkinVars["eggette"] = {
				special = EggAttacks
			}
		end
	else
		for player in players.iterate do
			if (player.mo)
			and not (player.mo.skin == "eggette")
			and (player.eggettebattle) then
				EggetteCleanup(player)
			end
		end
	end
end

local function BattleThink()
	BattleModSetup()
end

addHook("ThinkFrame", BattleThink)




/*
TAILS DOLL
  _______       _____ _       _____   _____   ____  _      _      
 |__   __|/\   |_   _| |     / ____| |  __ \ / __ \| |    | |     
    | |  /  \    | | | |    | (___   | |  | | |  | | |    | |     
    | | / /\ \   | | | |     \___ \  | |  | | |  | | |    | |     
    | |/ ____ \ _| |_| |____ ____) | | |__| | |__| | |____| |____ 
    |_/_/    \_\_____|______|_____/  |_____/ \____/|______|______|
                                                                  
                                                                  
*/

local function TailsSearch(actor)

	local dist = 620*FRACUNIT
	
	for player in players.iterate do
	
		if (player.spectator)
			continue
		end
		
		if (G_GametypeHasTeams())
			if (actor.tracer)
			and (CBW_Battle.MyTeam(actor.tracer, player))
			or (actor.team == player.ctfteam)
				continue
			end
		end

		if (player.pflags & PF_INVIS)
			continue // ignore notarget
		end

		if not (player.mo)
			continue
		end
		
		if not (player.mo.flags & MF_SHOOTABLE) then
			continue
		end

		if (player.mo.health <= 0)
			continue // dead
		end
		
		if (actor.tracer == player.mo) then
			continue
		end

		if (dist > 0)
		and (P_AproxDistance(P_AproxDistance(player.mo.x - actor.x, player.mo.y - actor.y), player.mo.z - actor.z) > dist)
			continue // Too far away
		end

		if not (P_CheckSight(actor, player.mo))
			continue // out of sight
		end
		
		actor.target = player.mo
		return true
	end

	actor.target = nil
	//return false;
end


local eggHurtTxt = {}

eggHurtTxt[MT_EGGS_TAILSDOLL] = "'s Tails Doll slapped "
eggHurtTxt[MT_JETTBULLET] = "'s Tails Doll shot "
eggHurtTxt[MT_EGGS_HOMINGBOMB] = "'s Amy Bomb exploded "

local function TailsDollDamageMsg(player, inf, src)
	local pName
	
	local hurtTxt = eggHurtTxt[inf.type]
	
	if (src)
	and (src.player)
		pName = TeamColorName(src.player.name, src.player)
	else
		if not (inf.valid) then return true end
		if (inf.pname) then
			pName = inf.pname
		end
	end
	
	if not (pName) then return true end
	
	print(pName..hurtTxt..TeamColorName(player.name, player))
	
	return true
end

addHook("HurtMsg", TailsDollDamageMsg, MT_EGGS_TAILSDOLL)
addHook("HurtMsg", TailsDollDamageMsg, MT_EGGS_HOMINGBOMB)


//Tails Doll Spawn script
//local function TailsDollCreate(mo)
local function TailsDollCreate(mo)
	mo.state = S_TAILSDOLL1
	mo.color = SKINCOLOR_ORANGE
	mo.exiting = false
	mo.pain = false
	mo.ai = 0
	mo.attacking = 0
	mo.attacktype = 0
	mo.ammo = 0
	mo.interval = 5
	mo.naturalcolor = mo.color
	mo.name = "\x87".."Tails Doll"
	//mo.name "Tails Doll"
	mo.telegraph = TICRATE
	mo.invisible = 0
	mo.cooldown = TICRATE*2
	mo.phase = 0
	mo.joining = TICRATE

	//asdfs
	mo.pain_tics = 0
	mo.hitcounter = 0
	
	local weight = 80
	local friction = 2
	local smooth = false
	local sentient = true
	
	//Apply values
	mo.weight = weight*FRACUNIT/100
	mo.momentum = (100-friction)*FRACUNIT/100
	mo.smooth = smooth
	mo.battle_atk = 0
	mo.battle_def = 0
	mo.pushtics = 0
	
	mo.ogx = mo.x
	mo.ogy = mo.y
	
end

//****
//Fuse and Thinker stuff

local joining = function(mo)
	//Gate
	if not (mo.joining > 0) then return false end
	mo.joining = $-1
	
	if not (leveltime & 3) then 
		P_SpawnMobj(mo.x,mo.y,mo.z,MT_IVSP)
	end
	
	if (mo.joining > 0) then
		mo.flags = $|MF_NOCLIPTHING
		mo.flags2 = $^^MF2_DONTDRAW
		mo.angle = $-ANG30
		mo.momz = mo.scale*2
		return true
	else
		mo.flags = $&~MF_NOCLIPTHING
		mo.flags2 = $&~MF2_DONTDRAW
		return false
	end
end

//General air control
local hover = function(mo,t)
	//Exiting gate
	if exiting then return end
	//Z coordinate to hover to
	local destz = mo.floorz
	if mo.ai and t then destz = max(mo.floorz,t.z) end //Get target z, but only if it exceeds floorz
	//Horizontal movement
	local spd = FixedHypot(mo.momx,mo.momy)
	local dir = R_PointToAngle2(0,0,mo.momx,mo.momy)
	P_Thrust(mo,dir,-spd/10) //Employ air friction
	//Vertical movement
	local water = CBW_Battle.WaterFactor(mo)
	local hoverthreshold = mo.scale*48
	local groundstop = mo.scale*10
	local thrustaccel = (mo.scale+1)/water
	local thrustmax = min(mo.scale*4,max(mo.scale*3,mo.scale*mo.hitcounter/4))
	local didthrust = false
	//Apply the "brakes" if we're about to hit the ground (prevents object from hitting a deathpit)
	if mo.z < destz +groundstop then
		mo.momz = max($,0)
		didthrust = true
	end
	//Rise if below hover threshold
	if mo.z < destz +hoverthreshold then
		mo.momz = min($+thrustaccel,thrustmax)
		didthrust = true
	end
	//Do aesthetic
	if didthrust and not(leveltime&3) and not(mo.invisible) then
		local dist = 8*mo.scale
		local x = mo.x+P_ReturnThrustX(mo,mo.angle+ANGLE_180,dist)
		local y = mo.y+P_ReturnThrustY(mo,mo.angle+ANGLE_180,dist)
		local z = mo.z+mo.height/2
		local spd = mo.scale*4
		local zspd = -mo.scale*6
		local d = P_SpawnMobj(x,y,z,MT_DUST)
		P_InstaThrust(d,mo.angle+ANGLE_180,mo.scale*4)
		P_SetObjectMomZ(d,zspd)
	end
end

//XY control
local zoning = function(mo,t)
	//Exiting gate
	if (mo.exiting) then return end
	//AI gate
	if not (mo.ai) then return end
	//Do "strafing"; every 128 tics the object will change direction
	local drift = 1
	if leveltime&128 then drift = -$ end
	
	//replace mo.hitcounter since it's never used
	local movespeed = 5
	
	//Acceleration increases as the Tails Doll takes more hits
	local water = CBW_Battle.WaterFactor(mo)
	local accel = mo.scale/30*movespeed/water
	
	//No target gate
	if not(t) then 
		if leveltime&64 then mo.angle = $+ANG10*drift end //Look around!
		
		if (R_PointToDist2(mo.x, mo.y, mo.ogx, mo.ogy) > 300*FRACUNIT) then
			P_Thrust(mo, R_PointToAngle2(mo.x, mo.y, mo.ogx, mo.ogy), accel)
		end
		return
	end
	
	local optimaldistmin = 256*mo.scale
	local optimaldistmax = 192*mo.scale
	local dist = R_PointToDist2(mo.x,mo.y,t.x,t.y)
	local angle = R_PointToAngle2(mo.x,mo.y,t.x,t.y)	
	//Turning
	local turn = ANG10
	local diff = angle-mo.angle
	if abs(diff) < turn then
		mo.angle = angle
	else
		mo.angle = $+max(-turn,min(turn,diff))
		angle = mo.angle
	end
	
	//stay within range of origin
	if (R_PointToDist2(mo.x, mo.y, mo.ogx, mo.ogy) <= 300*FRACUNIT) then
		//Apply acceleration
		if dist > optimaldistmin then //Get closer
			P_Thrust(mo,angle+ANGLE_45*drift,accel)
		elseif dist < optimaldistmax then //Move farther away
			P_Thrust(mo,angle+ANGLE_135*drift,accel)
		else //Full strafing
			P_Thrust(mo,angle+ANGLE_90*drift,accel)
		end
	else
		P_Thrust(mo, R_PointToAngle2(mo.x, mo.y, mo.ogx, mo.ogy), accel)
	end
end

//Exit animation script
local exiting = function(mo)
	if not(mo.exiting) then return false end
	//Do sparkle
	if not(leveltime&3) then P_SpawnMobj(mo.x,mo.y,mo.z,MT_IVSP) end
	mo.state = S_TAILSDOLL1
	mo.colorized = false
	mo.color = mo.naturalcolor
	//Flag object for stage exit
	mo.flags = $|MF_NOGRAVITY|MF_NOCLIPHEIGHT
	//Do blink
	mo.flags2 = $^^(MF2_DONTDRAW)
	//Spin
	mo.angle = $+ANG10
	//Halt xy momentum
	P_InstaThrust(mo,mo.angle,0)
	//Rise
	mo.momz = $+mo.scale/4
	//Once we've reached past the ceiling, the exit phase is complete and the object can be destroyed
	if mo.z > mo.ceilingz then
		P_RemoveMobj(mo)
	end
	return true
end

//Attack stuff
local headbullet = function(mo,t)
	local dist = mo.scale*4
	local height = mo.height
	local x = mo.x+P_ReturnThrustX(mo,mo.angle,dist)
	local y = mo.y+P_ReturnThrustY(mo,mo.angle,dist)
	local z = mo.z+height/2
	local bul = P_SpawnXYZMissile(mo.tracer,t,MT_JETTBULLET,x,y,z)
	
	if (bul) then
		//bul.tracer = mo.tracer
	end
	
	S_StartSound(mo,sfx_s3k4d)
end

local bombheight = FRACUNIT*256
local bombcompensate = FRACUNIT*64
local dangersparkle = function(mo,lerp)
	local h = FixedMul(bombheight,mo.scale)
	local c = FixedMul(bombcompensate,mo.scale)
	local height = min(mo.z+h,mo.ceilingz-c)
	height = CBW_Battle.FixedLerp(mo.z,$,lerp)
	local i = P_SpawnMobj(mo.x,mo.y,height,MT_IVSP)
	i.colorized = true
	i.color = mo.naturalcolor
end

local telebomb = function(mo,t)
	local h = FixedMul(bombheight,mo.scale)
	local c = FixedMul(bombcompensate,mo.scale)
	local height = min(mo.z+h,mo.ceilingz-c)
	local x = mo.x
	local y = mo.y
	local z = height
	P_SpawnXYZMissile(mo,t,MT_FBOMB,x,y,z)
	S_StartSound(mo,sfx_s3k51)
	dangersparkle(mo,FRACUNIT)
end

local attacktype = function(mo,nosound)
	//Bullets
	//if mo.ai == 2 then 
	mo.attacktype = 0
	mo.ammo = 3
	mo.interval = TICRATE/3
	//Get Sound
	if(nosound) then return end
	if mo.attacktype == 0 then S_StartSound(mo,sfx_cdfm60) end
	if mo.attacktype == 1 then S_StartSound(mo,sfx_s3k73) end
end

local attack = function(mo,t)
	//pain gate
	if mo.pain then
		mo.attacking = 0 //Reset attack state
		mo.fuse = 0
	return end
	//ai gate
	//if mo.ai < 2 then return end
	//Fuse not set?
	if mo.attacking == 0 and mo.fuse <= 0 then
		mo.fuse = TICRATE
	end
	//Tele-Bomb telegraph
	if mo.attacking == 1 and mo.attacktype == 1 and not(leveltime&3) then
		dangersparkle(mo,-FRACUNIT*(mo.fuse-mo.telegraph)/mo.telegraph)
	end
end

//Invisibility
local ghost = function(mo)
	mo.phase = $^^1 //Toggle phasing animation
	if not(mo.phase) then
		mo.invisible = $^^1 //Toggle invisiblity
	else //Start disappear
		mo.flags = $|MF_NOCLIPTHING
		if not(mo.invisible)
			S_StartSound(mo,sfx_s3k92)
		else //Reappear
			S_StartSound(mo,sfx_s3k8a)
			mo.attacking = 1
			attacktype(mo,true) //Prep attack type
		end
	end
	
	//Ambush attack
	if not(mo.invisible or mo.phase) then
		mo.flags = $&~MF_NOCLIPTHING
		return false //Continue fuse instructions
	else //Set next action time
		if mo.phase then
			mo.fuse = mo.telegraph
		else
			mo.fuse = mo.cooldown
		end
		return true //Gate off the fuse instructions
	end
end

//Draw control
local tdsprites = function(mo)
	if mo.pain_tics&1 then return end //pain flashing takes precedent
	if (mo.phase and leveltime&1) then
		mo.flags2 = $|MF2_DONTDRAW
	elseif not(mo.phase) and mo.invisible then
		mo.flags2 = $|MF2_DONTDRAW
	else
		mo.flags2 = $&~MF2_DONTDRAW
	end
	//Color
	mo.colorized = (mo.attacking == 1)
	if mo.colorized then
		mo.color = CBW_Battle.Choose(SKINCOLOR_WHITE,SKINCOLOR_RED,SKINCOLOR_YELLOW,SKINCOLOR_ORANGE)
	else
		mo.color = mo.naturalcolor
		
		if (mo.tracer)
		and (mo.tracer.valid) then
			mo.color = mo.tracer.color
		end
	
	end

	if mo.pain then
		mo.state = S_TAILSDOLL2
	elseif mo.state == S_TAILSDOLL2 then
		mo.state = S_TAILSDOLL1
	end
	
	return mo.state
end

//State/vars control
local tdstate = function(mo)
	//Regulate attack vars
	if mo.pain then
		mo.fuse = 0
	end
	//AI difficulty
	if mo.fuse == 0 then
		mo.attacking = 0 //Stationary
	end
	if mo.pain then return end
	mo.ai = 2
end

//Skin color
local color = function(mo,n)
	if not(n) then return end
	local list = {
		SKINCOLOR_ORANGE, //1 
		SKINCOLOR_SUNSET, //2 
		SKINCOLOR_RUST, //3
		SKINCOLOR_RED, //4 
		SKINCOLOR_PURPLE, //5 
		CBW_Battle.FlashColor(SKINCOLOR_SUPERORANGE1,SKINCOLOR_SUPERORANGE5), //6 
	}
	mo.naturalcolor = list[n]
end

//Target finding

local gettarget = function(mo)
	TailsSearch(mo, true, 620*mo.scale)
end

//stripped version of bashable thinker
local function TailsBashThinker(mo)
	if mo.pain_tics then
		mo.pain_tics = $-1
		if mo.pain_tics&1 then 
			mo.flags2 = $|MF2_DONTDRAW
		else
			mo.flags2 = $&~MF2_DONTDRAW
		end
	end
	
	if (mo.target) and not(mo.target.player) then //Bashables should only have player objects as their target.
		//Pushables will set other pushables as their target on collision, and this can cause problems
		if mo.target.target and mo.target.target.player then //But, if our target's target is a player object, due to a chain reaction...
			mo.target = mo.target.target //...Then we may as well borrow that target, so players receive proper credit for chain reactions
		else //Otherwise, we'll just clear our target value
			mo.target = nil
		end
	end
	//Do physics
	local trueweight = FixedMul(mo.weight,mo.scale) //Weight must reflect object scale
	local speed = FixedHypot(mo.momx,mo.momy) //Get current movement speed
	local threshold = trueweight*6 //Speed necessary before the object's velocity is considered "threatening"
	//Ball physics
	if mo.smooth then
		mo.friction = mo.momentum //Friction reflects "push momentum" at all times
		//Slope physics
		if (mo.standingslope)
			then
			local slope = mo.standingslope
			local m = FixedDiv(slope.zdelta,trueweight)
			P_Thrust(mo,slope.xydirection,m)
		end
	end
	if mo.pushed_last and not(mo.battle_atk) then
		mo.pushtics = $+1
	end
	//Object is moving at high speeds
	if speed/*+abs(mo.momz)*/ > threshold and(not(mo.sentient) or mo.pushed_last) then
		//Object can deal damage to other players
		mo.battle_atk = 1
		mo.friction = mo.momentum
		//Spawn thok mobj
		local t = P_SpawnMobj(mo.x,mo.y,mo.z,MT_THOK)
		if not(mo.target and mo.target.player) then
			t.color = SKINCOLOR_GREY
		else
			t.color = mo.target.player.skincolor
		end
		t.flags2 = $|MF2_SHADOW
		t.scale = FixedMul(FixedDiv(mo.height,t.height)*3/2,mo.scale) //Rescale to match object height
		//Object rotation is a visual indicator of movespeed
		mo.angle = $+FixedMul(speed,trueweight)/FRACUNIT*ANG1
		mo.pushtics = $+speed/FRACUNIT
		//Sliding FX
		if not(mo.smooth) and P_IsObjectOnGround(mo) and mo.pushtics >= 32
			mo.pushtics = 0
			S_StartSound(mo,sfx_s3k47)
			//Do dust particles
			local r = function()
				local w = mo.radius/FRACUNIT
				return P_RandomRange(-w,w)*FRACUNIT
			end
			P_SpawnMobj(mo.x+r(),mo.y+r(),mo.z,MT_DUST)
		end
	elseif P_IsObjectOnGround(mo) or mo.pushtics > 15 //Object is idle or not moving quickly. Keep our slate clean
		mo.pushed = nil
		mo.pushed_last = nil
		if not(mo.sentient) then
			mo.target = nil
		end
		mo.pushtics = 0
		mo.battle_atk = 0
		mo.flags = mo.info.flags
		mo.pain = false
	end
	
end


//prevent targetting at friendly tails dolls
local function TailsDollPreventFriendlyAim(actor)
	if not (consoleplayer) then return end
	
	local player = consoleplayer

	if (G_GametypeHasTeams())
		if (actor.tracer)
		and (CBW_Battle.MyTeam(actor.tracer, player))
		or (actor.team == player.ctfteam)
			actor.flags2 = $1|MF2_INVERTAIMABLE
		else
			actor.flags2 = $1 & ~MF2_INVERTAIMABLE
		end
	end
end

//Thinker body
local function TailsDollThinker(mo)

	if not (mo.health) then return end

	TailsDollPreventFriendlyAim(mo)

	mo.battle_atk = 0
	mo.battle_def = 0
	//color(mo, mo.ai)
	
	if (mo.tracer)
	and (mo.tracer.valid)
		mo.color = mo.tracer.color
	end
	
	if joining(mo) then return end
	if exiting(mo) then return end
	
	if (mo.state == S_INVISIBLE) then
		mo.state = S_TAILSDOLL2
	end
	
	//if not (mo.pain) then
	if not (mo.pushtics) then
		mo.battleobject = false
	end

	TailsBashThinker(mo)
	
	mo.ai = 2
	tdstate(mo)
	tdsprites(mo)
-- 	if mo.pain_tics then return end
	if mo.pain then return end
	gettarget(mo)
	hover(mo,mo.target)
	zoning(mo,mo.target)
	attack(mo,mo.target)
end

//Fuse body
local function TailsDollFuse(mo,t)
	//bashed gate
	if mo.pain then return end

	//target gate
	if not (t and t.valid) then return end
	//Start attack telegraph
	if mo.attacking == 0 then
		mo.fuse = mo.telegraph
		mo.attacking = 1
		attacktype(mo)
	return end
	
	
	//Launch attack
	if mo.attacking == 1 then
		mo.attacking = 2
	end
	if mo.attacktype == 0 then headbullet(mo,t) end
	if mo.attacktype == 1 then telebomb(mo,t) end
	mo.ammo = $-1
	//Multishot
	if mo.ammo then
		mo.fuse = mo.interval
	else //Restart the cycle
		mo.attacking = 0
		mo.fuse = TICRATE*2
	end
end


//prevent friendly fire
local function TailsDollHurtCheck(tails, tmo)

	if (tails.tracer)
	and (tails.tracer.valid) then
	
		if (tmo.flags & MF_MISSILE)
		and (tmo.target)
		and (tails.tracer)
			if (CBW_Battle.MyTeam(tmo.target, tmo))
			or (tmo.target == tails.tracer)
				return false
			end
		end
	
		if (tails.tracer == tmo)
		or (G_GametypeHasTeams())
		and (CBW_Battle.MyTeam(tails.tracer, tmo))
			return false
		end
	end
end
addHook("MobjCollide", TailsDollHurtCheck, MT_EGGS_TAILSDOLL)
addHook("MobjMoveCollide", TailsDollHurtCheck, MT_EGGS_TAILSDOLL)

addHook("MobjCollide", TailsDollHurtCheck, MT_JETTBULLET)
addHook("MobjMoveCollide", TailsDollHurtCheck, MT_JETTBULLET)

addHook("MobjSpawn", TailsDollCreate, MT_EGGS_TAILSDOLL)
addHook("MobjThinker", TailsDollThinker, MT_EGGS_TAILSDOLL)

local function TailsDollNoPoints(mo, inf, src)
	if (inf)
	or (src) then
		P_KillMobj(mo)
		return true
	end
end
addHook("MobjDeath", TailsDollNoPoints, MT_EGGS_TAILSDOLL)


local function TailsDollHit(td, inf, src)

	if (td.pain)
	or (td.pain_tics) then
		td.battleobject = true
		CBW_Battle.BashableCollision(td, source)
		return true
	end

	local source = src
	if (inf) then
		source = inf
	end
	
	td.battleobject = true
	CBW_Battle.BashableCollision(td, source)
	
	td.pain_tics = 15
	td.attacking = 0
	td.pain = true
	td.attacktype = 0
	
	td.health = $1-1
	
	if (td.health < 1) then
		P_KillMobj(td) //no points
		return false
	end
	
	td.battleobject = false
	
	return true
end

addHook("MobjDamage", TailsDollHit, MT_EGGS_TAILSDOLL)
addHook("MobjFuse", function(mo)
	TailsDollFuse(mo, mo.target)
	return true
end,
MT_EGGS_TAILSDOLL)
